import functools
import re
from typing import List, Optional


class Column:
    def __init__(self, index: int, name: Optional[str], values: List):
        """
        :param index: index of the column in the original table
        :param name: name of the column, None mean the column doesn't have any name (different from having empty name)
        :param values: values in each row
        """
        self.index = index
        self.name = name
        self.values = values

    @functools.cached_property
    def clean_name(self) -> Optional[str]:
        """Clean the name that may contain many unncessary spaces"""
        if self.name is None:
            return None
        return re.sub(r"\s+", " ", self.name).strip()

    def __getitem__(self, item):
        return self.values[item]

    def __setitem__(self, key, value):
        self.values[key] = value



import glob
import os, networkx as nx
import shutil
from pathlib import Path
from typing import Callable, Optional, Set, Tuple, List

import orjson


def viz_graph(g: nx.MultiDiGraph,
              node_fn: Callable[[str, dict], dict] = None,
              edge_fn: Callable[[str, dict], dict] = None,
              outdir: Optional[str] = None,
              graph_id: str = "default"):
    nodes = []
    edges = []

    if node_fn is None:
        node_fn = lambda uid, udata: {"label": uid}
    if edge_fn is None:
        edge_fn = lambda eid, edata: {"label": eid}

    for uid, vid, e, edata in g.edges(data=True, keys=True):
        edge = {
            "source": uid,
            "target": vid,
        }
        edge.update(edge_fn(e, edata))
        edges.append(edge)

    for uid in g:
        u = g.nodes[uid]
        node = {
            "id": uid,
        }
        node.update(node_fn(uid, u))
        nodes.append(node)

    return viz_raw_graph(nodes, edges, outdir, graph_id)
    

def viz_raw_graph(nodes: List[dict], edges: List[dict],
                  outdir: Optional[str], graph_id: str = "default"):
    if outdir is None:
        # directly return the result for rendering rather than plot them
        return nodes, edges

    Path(outdir).mkdir(exist_ok=True, parents=True)
    html_file = os.path.join(os.path.dirname(os.path.abspath(__file__)), "graph.html")
    shutil.copy(html_file, os.path.join(outdir, "graph.html"))

    with open(os.path.join(outdir, f"graph.{graph_id}.js"), "wb") as f:
        f.write(b"window.data = ")
        f.write(orjson.dumps({
            "nodes": nodes,
            "edges": edges
        }))

    graphs = sorted([
        Path(file).name[len("graph."):-3]
        for file in glob.glob(os.path.join(outdir, "graph.*.js"))
    ])
    with open(os.path.join(outdir, f"metadata.js"), "wb") as f:
        f.write(b"const metadata = ")
        f.write(orjson.dumps({
            "graphs": graphs,
            "currentGraph": [i for i, graph in enumerate(graphs) if graph == graph_id][0]
        }))



from typing import List, Dict, Tuple, Callable, Any, Optional

from sm.misc.graph import *


def node_fn(nid, ndata):
    colors = {
        "context": dict(fill="#C6E5FF", stroke="#5B8FF9"),
        "statement": dict(fill="#d9d9d9", stroke="#434343"),
        "qnode": dict(fill="#b7eb8f", stroke="#135200"),
        "column": dict(fill='#ffd666', stroke='#874d00')
    }

    if ndata['data']['column'] is not None:
        nodetype = 'column'
    elif ndata['data']['is_context']:
        nodetype = 'context'
    else:
        nodetype = 'statement' if nid.startswith("stmt") else "qnode"

    return {
        "label": ndata['data']['label'],
        "style": colors[nodetype],
        "labelCfg": {
            "style": {
                "fill": "black",
                "background": {
                    "padding": [4, 4, 4, 4],
                    "radius": 3,
                    **colors[nodetype]
                }
            }
        },
    }


# g = load_json(HOME_DIR / "graph_viz/graph.json")

# B = QueryBuilder()
#
# c1 = B.n("city", B.c("data.column", "=", 1))
# c2 = B.n("county", B.c("data.column", "=", 2))
# # c1 \
# #     .to(B.n(), B.c("data.predicate", "=", "P131")) \
# #     .to(c2, B.c("data.predicate", "=", "P131")) \
# #     .to(B.n(), B.c("data.predicate", "=", "P36")) \
# #     .to(c1, B.c("data.predicate", "=", "P36"))
# c1.to(B.n()).to(c2).to(B.n()).to(c1)
#
# g = query_graph(g, B.get_query())
# dump_json(g, HOME_DIR / "graph_viz/graph.2.json")
# viz_graph(g, node_fn, lambda eid, e: {"label": e['data']['predicate'] + f" #{e['data']['weight']}"}, HOME_DIR / "graph_viz/test", "origin")


def maximum_spanning_tree_solver(origin_sm_g: nx.MultiDiGraph):
    sm_g = nx.MultiDiGraph()
    for uid, vid, eid, edata in origin_sm_g.edges(data=True, keys=True):
        edata['weight'] = edata['data']['weight']
    resp = nx.algorithms.tree.branchings.maximum_spanning_arborescence(origin_sm_g, attr='weight',
                                                                       preserve_attrs=True)
    count = 0
    for s, t, e, edata in resp.edges(data=True, keys=True):
        edge = edata['data']
        if s not in sm_g:
            sm_g.add_node(s, **origin_sm_g.nodes[s])
        if t not in sm_g:
            sm_g.add_node(t, **origin_sm_g.nodes[t])
        sm_g.add_edge(s, t, key=edge['predicate'], data=edge)
        count += edge['weight']

    while True:
        remove_nodes = []
        for nid in sm_g:
            n = sm_g.nodes[nid]['data']
            if not n['column'] is not None and not n['is_context']:
                if sm_g.out_degree(nid) == 0:
                    remove_nodes.append(nid)

        for nid in remove_nodes:
            sm_g.remove_node(nid)

        if len(remove_nodes) == 0:
            break
    return sm_g

# g2 = maximum_spanning_tree_solver(g)
# viz_graph(g2, node_fn, lambda eid, e: {"label": e['data']['predicate'] + f" #{e['data']['weight']}"}, HOME_DIR / "graph_viz/test", "pruning")


from sm.misc.fn_cache import *
from sm.misc.deser import *
from sm.misc.funcs import *
from sm.misc.exp_manager import *
from sm.misc.logger import *
from sm.misc.bijection import Bijection
from sm.misc.matrix import Matrix


class UnreachableError(Exception):
    pass



from dataclasses import dataclass
from pathlib import Path
from zipfile import ZipFile
from typing import List, TypeVar, Generic, Union, Callable
import orjson

from sm.misc import deserialize_json, get_latest_path
from sm.outputs import SemanticModel

T = TypeVar("T")


@dataclass
class Example(Generic[T]):
    sms: List[SemanticModel]
    table: T


def load(
    data_dir: Union[str, Path], table_deser: Callable[[dict], T]
) -> List[Example[T]]:
    """Load dataset from a folder. Assuming the following structure:

    descriptions (containing semantic descriptions of tables)
    ├── <table_fs_id>
    │   ├── version.01.json
    │   ├── version.02.json
    │   └── ...
        or
    ├── <table_fs_id>.json
    ├── ...
    tables (containing list of tables, the type of table depends on )
    ├── <table_fs_id>.json
    ├── ...

    We also support compressing formats such as .zip.
    descriptions
    ├── part-<num>.zip (no nested version folders)
    │   ├── <table_fs_id>.json
    |   |   ...
    tables
    ├── part-<num>.zip
    │   ├── <table_fs_id>.json
    |   |   ...

    Args:
        data_dir:
        table_deser: deserialize the table from dictionary

    Returns:

    """
    data_dir = Path(data_dir)
    examples = []
    for infile in sorted((data_dir / "tables").iterdir()):
        if infile.name.startswith("."):
            continue
        if infile.name.endswith(".json"):
            example_id = infile.stem

            table = table_deser(deserialize_json(infile))

            if (data_dir / f"descriptions/{example_id}").exists():
                desc_file = get_latest_path(
                    data_dir / f"descriptions/{example_id}/version.json"
                )
                assert desc_file is not None
            else:
                desc_file = data_dir / f"descriptions/{example_id}.json"
                assert desc_file.exists()

            raw_sms = deserialize_json(desc_file)
            sms = [SemanticModel.from_dict(sm) for sm in raw_sms]

            examples.append(Example(sms=sms, table=table))
        elif infile.name.endswith(".zip"):

            part = {}
            with ZipFile(infile, mode="r") as zf:
                for file in zf.infolist():
                    if not file.filename.endswith(".json"):
                        continue

                    table_id = Path(file.filename).stem
                    with zf.open(file, mode="r") as f:
                        table = table_deser(orjson.loads(f.read()))
                    part[table_id] = table

            lst = []
            with ZipFile(data_dir / "descriptions" / infile.name, mode="r") as zf:
                for file in zf.infolist():
                    table_id = Path(file.filename).stem
                    if table_id not in part:
                        continue

                    assert file.filename.endswith(".json")
                    with zf.open(file, mode="r") as f:
                        sms = [
                            SemanticModel.from_dict(sm) for sm in orjson.loads(f.read())
                        ]
                        lst.append(Example(sms=sms, table=part[table_id]))

            assert len(lst) == len(part)
            examples.extend(lst)

    return examples



from typing import Generic, TypeVar, Dict
from typing_extensions import Self


V1 = TypeVar("V1")
V2 = TypeVar("V2")


class Bijection(Generic[V1, V2]):
    def __init__(self):
        self.x2prime: Dict[V1, V2] = {}
        self.prime2x: Dict[V2, V1] = {}

    def size(self) -> int:
        return len(self.x2prime)

    def add(self, x: V1, xprime: V2) -> bool:
        """Add a invertible mapping between x and xprime. Return true if
        successful (even when an existing mapping exists), false if x or xprime is already mapped to another value.
        """
        if (x in self.x2prime) != (xprime in self.prime2x):
            return False

        if x in self.x2prime:
            return self.x2prime[x] == xprime

        self.x2prime[x] = xprime
        self.prime2x[xprime] = x

        return True

    def check_add(self, x: V1, xprime: V2) -> Self:
        """Add a invertible mapping between x and xprime. Return Self if
        successful or raise ValueError exception if x or xprime is already mapped to another value.
        """
        flag = self.add(x, xprime)
        if not flag:
            raise ValueError(
                f"Cannot add mapping {x} <-> {xprime} because it contradicts with existing mapping {self.x2prime.get(x, '∅')} -> {self.prime2x.get(xprime, '∅')}"
            )
        return self



import math
from typing import Dict, Callable, List, Mapping, Set, Iterable, Tuple
from dataclasses import dataclass
from sm.evaluation.sm_metrics import ScoringFn
from tqdm import tqdm


ItemParents = Dict[str, Iterable[str]]


@dataclass
class HierarchyScoringFn(ScoringFn):
    # mapping from the item to its parents and their distances
    # 1 is direct parent, 2 is grandparent, etc.
    item2parents: Mapping[str, Mapping[str, int]]

    @staticmethod
    def construct(
        items: List[str],
        get_item_parents: Callable[[str], Iterable[str]],
        get_item_uri: Callable[[str], str],
        verbose: bool = False,
    ):
        """Create a HierarchyScoringFn from a list of items

        Args:
            items: List of items to build the scoring function
            get_item_parents: Function that returns the parents of an item
            get_item_uri: Function that returns the URI of an item
        """
        item2parents = {item: {} for item in items}

        for item in tqdm(items, disable=not verbose):
            # a mapping of visited item to the distance that we encounter it
            visited = {}
            stack: List[Tuple[str, int]] = [
                (parent, 1) for parent in get_item_parents(item)
            ]
            while len(stack) > 0:
                node, distance = stack.pop()
                if node in visited and distance >= visited[node]:
                    # we have visited this node before and since last time we visit
                    # the previous route is shorter, so we don't need to visit it again
                    continue

                visited[node] = distance
                item2parents[item][node] = min(
                    distance, item2parents[item].get(node, float("inf"))
                )

                for parent in get_item_parents(node):
                    stack.append((parent, distance + 1))

        return HierarchyScoringFn(
            {
                get_item_uri(k): {get_item_uri(k2): v2 for k2, v2 in v.items()}
                for k, v in item2parents.items()
            }
        )

    def get_match_score(self, pred_item: str, target_item: str):
        if pred_item == target_item:
            return 1.0
        if pred_item in self.item2parents[target_item]:
            # pred_predicate is the parent of the target
            distance = self.item2parents[target_item][pred_item]
            if distance > 5:
                return 0.0
            return math.pow(0.8, distance)
        if target_item in self.item2parents[pred_item]:
            distance = self.item2parents[pred_item][target_item]
            if distance > 3:
                return 0.0
            return math.pow(0.7, distance)
        return 0.0



from __future__ import annotations
from typing import Dict, Union
from pathlib import Path
from dataclasses import dataclass
from sm.misc.deser import deserialize_yml

from sm.namespaces.prefix_index import PrefixIndex


class OutOfNamespace(Exception):
    pass


default_ns_file = Path(__file__).absolute().parent.parent / "data/namespaces.yml"


@dataclass
class Namespace:
    """A helper class for converting between absolute URI and relative URI."""

    __slots__ = ("prefix2ns", "prefix_index")

    prefix2ns: Dict[str, str]
    prefix_index: PrefixIndex

    @classmethod
    def from_file(cls, infile: Union[Path, str] = default_ns_file):
        prefix2ns = dict(deserialize_yml(infile))
        ns2prefix = {v: k for k, v in prefix2ns.items()}
        assert len(ns2prefix) == len(prefix2ns), "Duplicated namespaces"
        prefix_index = PrefixIndex.create(ns2prefix)

        return cls(prefix2ns=prefix2ns, prefix_index=prefix_index)

    def get_abs_uri(self, rel_uri: str):
        """Get absolute URI from relative URI."""
        prefix, name = rel_uri.split(":", 2)
        return self.prefix2ns[prefix] + name

    def get_rel_uri(self, abs_uri: str):
        """Get relative URI from absolute URI."""
        prefix = self.prefix_index.get(abs_uri)
        if prefix is None:
            raise OutOfNamespace(
                f"Cannot simply the uri `{abs_uri}` as its namespace is not defined"
            )

        return f"{prefix}:{abs_uri.replace(self.prefix2ns[prefix], '')}"

    def is_rel_uri(self, uri: str):
        """Check if an URI is relative."""
        return uri.count(":") == 1

    def is_uri_in_ns(self, abs_uri: str, prefix: str):
        """Check if an absolute URI is in a namespace specified by the prefix."""
        return abs_uri.startswith(self.prefix2ns[prefix])

    def get_resource_id(self, abs_uri: str):
        """
        Get the resource id from an absolute URI in its namespace, stripped out the namespace prefix.
        There is no guarantee that resources in different namespaces won't have the same id.

        Examples:
        - http://www.wikidata.org/entity/Q512 -> Q512
        - http://dbpedia.org/resource/Berlin -> Berlin
        """
        prefix = self.prefix_index.get(abs_uri)
        if prefix is None:
            raise OutOfNamespace(
                f"Cannot get resource id of the uri `{abs_uri}` as its namespace is not defined"
            )
        return abs_uri.replace(self.prefix2ns[prefix], "")

    def is_compatible(self, ns: Namespace) -> bool:
        """Test if prefixes of two namespaces are the same"""
        return all(
            self.prefix2ns[prefix] == ns.prefix2ns[prefix]
            for prefix in set(self.prefix2ns.keys()).intersection(ns.prefix2ns.keys())
        )



import glob
from inspect import signature
from multiprocessing import get_context
import re
import math
from contextlib import contextmanager
from multiprocessing.pool import Pool, ThreadPool
from operator import itemgetter
from pathlib import Path
from typing import (
    Dict,
    Iterable,
    Literal,
    TypeVar,
    Union,
    Callable,
    Any,
    List,
    Optional,
    KeysView,
)

from loguru import logger
from tqdm.auto import tqdm
import importlib


K = TypeVar("K")
V = TypeVar("V")


def str2bool(x):
    assert x in {"True", "False", "true", "false", "null"}
    if x == "null":
        return None
    return x.lower() == "true"


def nullable_str(x):
    if x == "null":
        return None
    return x


def str2int(x):
    if x == "null":
        return None
    return int(x)


def assert_not_null(x: Optional[V]) -> V:
    assert x is not None
    return x


def percentage(a: Union[float, int], b: Union[float, int]) -> str:
    return "%.2f%% (%d/%d)" % (a * 100 / b, a, b)


def filter_duplication(lst: List[Any], key_fn: Optional[Callable[[Any], Any]] = None):
    key_fn = key_fn or identity_func
    keys = set()
    new_lst = []
    for item in lst:
        k = key_fn(item)
        if k in keys:
            continue

        keys.add(k)
        new_lst.append(item)
    return new_lst


def identity_func(x):
    return x


def get_latest_version(file_pattern: Union[str, Path]) -> int:
    """Assuming the file pattern select list of files tagged with an integer version for every run, this
    function return the latest version number that you can use to name your next run.

    For example:
    1. If your pattern matches folders: version_1, version_5, version_6, this function will return 6.
    2. If your pattern does not match anything, return 0
    """
    files = [Path(file) for file in sorted(glob.glob(str(file_pattern)))]
    if len(files) == 0:
        return 0

    file = sorted(files)[-1]
    match = re.match("[^0-9]*(\d+)[^0-9]*", file.name)
    if match is None:
        raise Exception("Invalid naming")
    return int(match.group(1))


def get_incremental_path(
    path: Union[str, Path],
    create_if_missing: bool = True,
    delimiter_char: Optional[str] = None,
) -> Path:
    """Get an incremental path.

    Example:
        >>> get_incremental_path("/test/data")
        '/test/data_01'
        >>> mkdir('/test/data_01')
        >>> # create a new folder with version 01, so the next time you call, it will be `data_02`
        >>> get_incremental_path("/test/data")
        '/test/data_02'

    Arguments:
        path: path to the folder you want to get incremental path
        create_if_missing:
            if the path has a suffix, it is treated as a file and if its parent directory does not exist, create the parent directory.
            if the path does not have a suffix, it is treated as a folder and create if it does not exist.
        delimiter_char: if you want to use a different delimiter, you can specify it here.
            Default it is _ for no suffix path (folder) and . for file
    """
    path = Path(str(path))
    if delimiter_char is None:
        if path.suffix == "":
            delimiter_char = "_"
        else:
            delimiter_char = "."

    pattern = path.parent / f"{path.stem}{delimiter_char}*{path.suffix}"
    version = get_latest_version(pattern) + 1

    newpath = path.parent / f"{path.stem}{delimiter_char}{version:02d}{path.suffix}"
    if create_if_missing:
        if path.suffix == "":
            newpath.mkdir(parents=True)
        else:
            newpath.parent.mkdir(parents=True, exist_ok=True)
    return newpath


def get_latest_path(path: Union[str, Path]) -> Optional[Path]:
    path = Path(str(path))
    pattern = path.parent / f"{path.stem}*{path.suffix}"
    version = get_latest_version(pattern)
    if version == 0:
        return None

    (match_path,) = list(path.parent.glob(f"{path.stem}*{version:02d}{path.suffix}"))
    return match_path


def auto_wrap(
    word: str,
    max_char_per_line: int,
    delimiters: Optional[List[str]] = None,
    camelcase_split: bool = True,
) -> str:
    """
    Treat this as optimization problem, where we trying to minimize the number of line break
    but also maximize the readability in each line, i.e: maximize the number of characters in each lines

    Using greedy search.
    :param word:
    :param max_char_per_line:
    :param delimiters:
    :return:
    """
    # split original word by the delimiters
    if delimiters is None:
        delimiters = [" ", ":", "_", "/"]

    sublines: List[str] = [""]
    for i, c in enumerate(word):
        if c not in delimiters:
            sublines[-1] += c

            if (
                camelcase_split
                and not c.isupper()
                and i + 1 < len(word)
                and word[i + 1].isupper()
            ):
                # camelcase_split
                sublines.append("")
        else:
            sublines[-1] += c
            sublines.append("")

    new_sublines: List[str] = [""]
    for line in sublines:
        if len(new_sublines[-1]) + len(line) <= max_char_per_line:
            new_sublines[-1] += line
        else:
            new_sublines.append(line)

    return "\n".join(new_sublines)


def flatten_dict(odict: dict, result: Optional[dict] = None, prefix: str = ""):
    if result is None:
        result = {}

    for k, v in odict.items():
        if isinstance(v, dict):
            flatten_dict(v, result, prefix=prefix + k + ".")
        else:
            result[prefix + k] = v
    return result


def flatten_list(lst: list) -> list:
    """Flatten nested list, anything that is instance of a list get flatten"""
    output = []
    for item in lst:
        if isinstance(item, list):
            for subitem in item:
                if isinstance(subitem, list):
                    output += flatten_list(subitem)
                else:
                    output.append(subitem)
        else:
            output.append(item)
    return output


def group_by(lst: Iterable[V], key: Callable[[V], K]) -> Dict[K, List[V]]:
    odict = {}
    for item in lst:
        k = key(item)
        if k not in odict:
            odict[k] = []
        odict[k].append(item)
    return odict


def create_group_by_index(
    lst: Iterable[V], key: Callable[[V], K]
) -> Dict[K, List[int]]:
    odict = {}
    for i, item in enumerate(lst):
        k = key(item)
        if k not in odict:
            odict[k] = []
        odict[k].append(i)
    return odict


def datasize(num, suffix="B"):
    """Get human friendly file size
    https://gist.github.com/cbwar/d2dfbc19b140bd599daccbe0fe925597#gistcomment-2845059

    Args:
        num (int): Bytes value
        suffix (str, optional): Unit. Defaults to 'B'.

    Returns:
        str: file size0
    """
    if num == 0:
        return "0"
    magnitude = int(math.floor(math.log(num, 1024)))
    val = num / math.pow(1024, magnitude)
    if magnitude > 7:
        return "{:3.1f}{}{}".format(val, "Yi", suffix)
    return "{:3.1f}{}{}".format(
        val, ["", "Ki", "Mi", "Gi", "Ti", "Pi", "Ei", "Zi"][magnitude], suffix
    )


class ParallelMapFnWrapper:
    def __init__(self, fn: Callable, ignore_error=False):
        self.fn = fn
        fn_params = signature(fn).parameters
        self.spread_fn_args = len(fn_params) > 1
        self.ignore_error = ignore_error

    def run(self, args):
        idx, r = args
        try:
            if self.spread_fn_args:
                r = self.fn(*r)
            else:
                r = self.fn(r)
            return idx, r
        except:
            logger.error(f"[ParallelMap] Error while process item {idx}")
            if self.ignore_error:
                return idx, None
            raise


def parallel_map(
    fn,
    inputs,
    show_progress=False,
    progress_desc="",
    is_parallel=True,
    use_threadpool=False,
    n_processes: Optional[int] = None,
    ignore_error: bool = False,
    start_method: Literal["fork", "spawn", "forkserver"] = "fork",
):
    if not is_parallel:
        iter = (fn(item) for item in inputs)
        if show_progress:
            iter = tqdm(iter, total=len(inputs), desc=progress_desc)
        return list(iter)

    if use_threadpool:
        with ThreadPool(processes=n_processes) as pool:
            iter = pool.imap_unordered(
                ParallelMapFnWrapper(fn, ignore_error).run, enumerate(inputs)
            )
            if show_progress:
                iter = tqdm(iter, total=len(inputs), desc=progress_desc)
            results = list(iter)
            results.sort(key=itemgetter(0))
    else:
        with get_context(start_method).Pool(processes=n_processes) as pool:
            iter = pool.imap_unordered(
                ParallelMapFnWrapper(fn, ignore_error).run, enumerate(inputs)
            )
            if show_progress:
                iter = tqdm(iter, total=len(inputs), desc=progress_desc)
            results = list(iter)
            results.sort(key=itemgetter(0))
    return [v for i, v in results]


@contextmanager
def print2file(file_path: Union[str, Path], mode="w", file_only: bool = False):
    """Yield a print function that can be both print to file or print to std"""
    Path(file_path).parent.mkdir(exist_ok=True, parents=True)
    origin_print = print
    with open(str(file_path), mode) as f:

        def print_fn(*args):
            if not file_only:
                origin_print(*args)
            origin_print(*args, file=f)

        try:
            yield print_fn
        finally:
            pass


K = TypeVar("K")
V = TypeVar("V")
V2 = TypeVar("V2")


class DictProxy(Dict[K, V2]):
    """Dictionary proxy to access objects' property

    Args:
        odict: dictionary of object
        access: function to access property of an object
    """

    def __init__(self, odict: Dict[K, V], access: Callable[[V], V2]):
        self.odict = odict
        self.access = access

    def __iter__(self):
        return self.odict.__iter__()

    def __getitem__(self, item):
        return self.access(self.odict[item])

    def __contains__(self, item):
        return item in self.odict

    def keys(self) -> KeysView[K]:
        return self.odict.keys()

    def values(self):
        return (self.access(v) for v in self.odict.values())

    def items(self):
        return ((k, self.access(v)) for k, v in self.odict.items())


def import_func(func_ident: str) -> Callable:
    """Import function from string, e.g., sm.misc.funcs.import_func"""
    lst = func_ident.rsplit(".", 2)
    if len(lst) == 2:
        module, func = lst
        cls = None
    else:
        module, cls, func = lst
        try:
            importlib.import_module(module + "." + cls)
            module = module + "." + cls
            cls = None
        except ModuleNotFoundError:
            pass

    module = importlib.import_module(module)
    if cls is not None:
        module = getattr(module, cls)

    return getattr(module, func)


def import_attr(attr_ident: str):
    lst = attr_ident.rsplit(".", 1)
    module, cls = lst
    module = importlib.import_module(module)
    return getattr(module, cls)


class Proxy:
    """Proxy object, can be used for delayed initialization"""

    def __init__(self, _object):
        self._object = _object

    def __getattr__(self, name):
        return getattr(self._object(), name)


R = TypeVar("R")
OBJECTS = {}


def get_instance(constructor: Callable[[], R], name: Optional[str] = None) -> R:
    """A utility function to get a singleton, which can be created from the given constructor.

    One use case of this function is we have a big object that is expensive to send
    to individual task repeatedly. If the process are retrieved from a pool,
    this allows us to create the object per process instead of per task.
    """
    global OBJECTS

    if name is None:
        assert (
            constructor.__name__ != "<lambda>"
        ), "Cannot use lambda as a name because it will keep changing"
        name = constructor  # type: ignore

    if name not in OBJECTS:
        logger.trace("Create a new instance of {}", name)
        OBJECTS[name] = constructor()
    return OBJECTS[name]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['sm',
 'sm.evaluation',
 'sm.inputs',
 'sm.misc',
 'sm.misc.graph',
 'sm.namespaces',
 'sm.outputs']

package_data = \
{'': ['*'], 'sm': ['data/*']}

install_requires = \
['chardet>=4.0.0,<5.0.0',
 'colorama>=0.4.4',
 'graph-wrapper>=1.4.0,<2.0.0',
 'ipython>=8.0.1,<9.0.0',
 'loguru>=0.6.0,<0.7.0',
 'lz4>=4.0.2,<5.0.0',
 'matplotlib>=3.5.3,<4.0.0',
 'orjson>=3.8.0,<4.0.0',
 'pandas>=1.4.1,<2.0.0',
 'pydot>=1.4.2,<2.0.0',
 'pyrsistent>=0.17.3,<0.18.0',
 'python-slugify>=6.1.2,<7.0.0',
 'rdflib>=6.1.1,<7.0.0',
 'ruamel.yaml>=0.17.9,<0.18.0',
 'tqdm>=4.64.0,<5.0.0',
 'ujson>=5.1.0,<6.0.0']

setup_kwargs = {
    'name': 'sem-desc',
    'version': '3.8.2',
    'description': 'Package providing basic functionalities for the semantic modeling problem',
    'long_description': '# SEM-DESC ![PyPI](https://img.shields.io/pypi/v/sem-desc)\n\nContaining basic functions (input, output, evaluation metrics) for the semantic modeling problem.\n',
    'author': 'Binh Vu',
    'author_email': 'binh@toan2.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/binh-vu/sm',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from __future__ import annotations
import threading
from abc import abstractmethod
from contextlib import contextmanager
from typing import Any, Optional, Callable


_container = threading.local()
_container.logger = None


class ContextLogger:
    @abstractmethod
    def log(self, ns: str, data: dict, mutual_data: dict):
        pass

    @abstractmethod
    def clear(self):
        pass


@contextmanager
def context_logger(
    context: dict,
    constructor: Optional[Callable[[dict], ContextLogger]] = None,
    disable: bool = False,
):
    global _container
    if disable:
        yield None
        return

    previous_logger = _container.logger
    try:
        _container.logger = (
            ContextLogger() if constructor is None else constructor(context)
        )
        yield _container.logger
    finally:
        if _container.logger is not None:
            _container.logger.clear()
        _container.logger = previous_logger


def global_logger(
    context: dict, constructor: Optional[Callable[[dict], ContextLogger]] = None
):
    global _container
    if _container.logger is not None:
        raise Exception("Can only register global logger before registering any logger")

    _container.logger = ContextLogger() if constructor is None else constructor(context)


def log(ns: str, **kwargs: Any):
    """Log objects. It accepts a special key: `mutual_data`, which is a dictionary
    of mutual objects. Every object in this dictionary will be deep copy to prevent any
    changes in the original object.
    """
    global _container

    if _container.logger is not None:
        if "mutual_data" in kwargs:
            mutual_data = kwargs.pop("mutual_data")
        else:
            mutual_data = {}
        _container.logger.log(ns, kwargs, mutual_data)
        return True
    return False



from pathlib import Path

import networkx as nx
from typing import List, Union, Optional

import ujson

from sm.misc.graph.query import PropPath
from sm.misc.deser import serialize_json, deserialize_json


def dump_json(g: nx.MultiDiGraph, outfile: Optional[str], node_props: List[str]=None, edge_props: List[str]=None, allow_unknown_key: bool=False):
    if node_props is not None:
        node_props = [PropPath(p, allow_unknown_key) for p in node_props]
    if edge_props is not None:
        edge_props = [PropPath(p, allow_unknown_key) for p in edge_props]

    nodes = []
    for nid, ndata in g.nodes(data=True):
        if node_props is not None:
            tmp = {}
            for p in node_props:
                p.copy_to_dict(nid, ndata, tmp)
        else:
            tmp = ndata
        nodes.append((nid, tmp))

    edges = []
    for uid, vid, eid, edata in g.edges(data=True, keys=True):
        if edge_props is not None:
            tmp = {}
            for p in edge_props:
                p.copy_to_dict(eid, edata, tmp)
        else:
            tmp = edata
        edges.append(((uid, vid, eid), tmp))

    if outfile is not None:
        serialize_json({
            "nodes": nodes,
            "edges": edges
        }, outfile, indent=4)
    else:
        return ujson.dumps({"nodes": nodes, "edges": edges})


def load_json(infile_or_data: Union[Path, str, dict, list]):
    if isinstance(infile_or_data, (list, dict)):
        data = infile_or_data
    else:
        data = deserialize_json(infile_or_data)
    g = nx.MultiDiGraph()
    for nid, ndata in data['nodes']:
        g.add_node(nid)

    for i in range(len(data['edges'])):
        data['edges'][i][0] = tuple(data['edges'][i][0])

    for (uid, vid, eid), edata in data['edges']:
        g.add_edge(uid, vid, eid)

    nx.set_node_attributes(g, dict(data['nodes']))
    nx.set_edge_attributes(g, dict(data['edges']))
    return g